"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _helperPluginUtils = require("@babel/helper-plugin-utils");

var _core = require("@babel/core");

var _parser = require("@babel/parser");

// Babel plugin for operator overloading
// This plugin changes withOperatorsFrom() calls into declarations
// of operator sets, and arithmetic operations into calls into the
// operator overloading runtime, if they take place within a
// withOperatorsFrom() scope.
// Note that the proposal uses a `with operators from` statement instead;
// this version doesn't use such new syntax to avoid the complexity
// that comes with a parser change.
// This plugin runs in a single visitor pass.
// The this object has the following properties included in it:
// {
//    shim: uid of the required shim module (undefined when stack empty)
//    stack: An Array of {
//      operators: uid of the current modules object
//      path: The path that owns this operator set
//    }
// }
// When the stack is empty, there's no operator overloading registered,
// and no transformation is done until entering a block with a
// withOperatorsFrom() call as a StatementExpressoin.
// The behavior for each element is as follows:
//   Blocks and programs:
//     - Check for a top-level withOperatorsFrom() statement. If found,
//       - Require the shim if empty, and add this as the first statement
//         in the block.
//       - Make the next statement a declaration of a uid for the operators
//         variable, initialized based on the outer operators variable
//         found from the stack.   let newuid = shim._declareOperators(olduid);
//       - Save the operators variable and this node on the stack
//       - As a post callback, pop the top of the stack if we pushed something.
//   Function calls:
//     - If the function is withOperatorsFrom:
//       - If the stack is empty, throw an error (was not at the top level).
//       - Otherwise, turn it into shim._withOperatorsFrom(operators, args...)
//   Operators:
//     - Replace all x= operators with the expanded var = var x arg form
//     - Replace all infix mathematical operators with calls to
//       _shim._binary(_operators)
//     - Replace unary operators with _shim._unary(_operators)
//     - Replace preincrement and postincrement operators with an
//       assignment and call to the pos or neg operators, with
//       _shim._unary(_operators)
var withOperatorsFromTemplate = (0, _core.template)("\n  SHIM._withOperatorsFrom(OPERATORS, ARGS)\n");
var requireShimTemplate = (0, _core.template)("\n  const SHIM = require(\"@littledan/operator-overloading-shim\");\n");
var declareOperatorsTemplate = (0, _core.template)("\n  const OPERATORS = SHIM._declareOperators(OUTER);\n");
var unaryOperatorTemplate = (0, _core.template)("\n  SHIM._unary(OPERATOR, EXPRESSION, OPERATORS)\n");
var binaryOperatorTemplate = (0, _core.template)("\n  SHIM._binary(OPERATOR, LEFT, RIGHT, OPERATORS)\n");
var preIncrementTemplate = (0, _core.template)("\n  EXPRESSION = SHIM._unary(OPERATOR, EXPRESSION, OPERATORS)\n");
var postIncrementTemplate = (0, _core.template)("\n  TEMPORARY = EXPRESSION,\n  EXPRESSION = SHIM._unary(OPERATOR, EXPRESSION, OPERATORS),\n  TEMPORARY\n");

function isWithOperatorsFrom(node) {
  return _core.types.isIdentifier(node.callee) && node.callee.name === "withOperatorsFrom";
}

var visitBlockStatementLike = {
  enter: function enter(path) {
    if (!path.node.body.some(function (statement) {
      return _core.types.isExpressionStatement(statement) && _core.types.isCallExpression(statement.expression) && isWithOperatorsFrom(statement.expression);
    })) return;
    var prelude = [];

    if (this.shim === undefined) {
      this.shim = path.scope.generateUidIdentifier("shim");
      prelude.push(requireShimTemplate({
        SHIM: this.shim
      }));
    }

    var operators = path.scope.generateUidIdentifier("operators");
    var outer = this.inactive() ? undefined : this.peek().operators;
    this.stack.push({
      operators: operators,
      path: path
    });
    prelude.push(declareOperatorsTemplate({
      OPERATORS: operators,
      SHIM: this.shim,
      OUTER: outer
    }));
    path.unshiftContainer('body', prelude);
  },
  exit: function exit(path) {
    if (this.peek().path === path) {
      this.stack.pop();
      if (this.inactive()) this.shim = undefined;
    }
  }
};
var fixedBinaryOperators = new Set(["===", "!==", "in", "instanceOf"]);

var _default = (0, _helperPluginUtils.declare)(function (api) {
  api.assertVersion(7);
  return {
    pre: function pre() {
      var _this = this;

      this.stack = [];

      this.peek = function () {
        return _this.stack[_this.stack.length - 1];
      };

      this.inactive = function () {
        return _this.stack.length === 0;
      };
    },
    post: function post() {
      if (!this.inactive() || this.shim !== undefined) {
        throw "internal error";
      }
    },
    visitor: {
      BlockStatement: visitBlockStatementLike,
      Program: visitBlockStatementLike,
      CallExpression: function CallExpression(path) {
        if (!isWithOperatorsFrom(path.node)) return;

        if (this.inactive()) {
          throw path.buildCodeFrameError("withOperatorsFrom calls must be statements, not nested expressions.");
        }

        var uid = this.peek().operators;
        path.replaceWith(withOperatorsFromTemplate({
          SHIM: this.shim,
          OPERATORS: uid,
          ARGS: path.node.arguments
        }));
      },
      UpdateExpression: function UpdateExpression(path) {
        if (this.inactive()) return;
        var statement;

        if (path.node.prefix) {
          statement = preIncrementTemplate({
            SHIM: this.shim,
            OPERATOR: _core.types.StringLiteral(path.node.operator),
            EXPRESSION: path.node.argument,
            OPERATORS: this.peek().operators
          });
        } else {
          var temporary = path.scope.generateUidIdentifier("t");
          statement = postIncrementTemplate({
            SHIM: this.shim,
            OPERATOR: _core.types.StringLiteral(path.node.operator),
            EXPRESSION: path.node.argument,
            OPERATORS: this.peek().operators,
            TEMPORARY: temporary
          });
        }

        path.replaceWith(statement);
      },
      UnaryExpression: function UnaryExpression(path) {
        if (this.inactive()) return;
        var operator = {
          "+": "pos",
          "-": "neg",
          "~": "~"
        }[path.node.operator];
        if (operator === undefined) return;
        path.replaceWith(unaryOperatorTemplate({
          SHIM: this.shim,
          OPERATOR: _core.types.StringLiteral(operator),
          EXPRESSION: path.node.argument,
          OPERATORS: this.peek().operators
        }));
      },
      BinaryExpression: function BinaryExpression(path) {
        if (this.inactive()) return;
        if (fixedBinaryOperators.has(path.node.operator)) return;
        path.replaceWith(binaryOperatorTemplate({
          SHIM: this.shim,
          OPERATOR: _core.types.StringLiteral(path.node.operator),
          LEFT: path.node.left,
          RIGHT: path.node.right,
          OPERATORS: this.peek().operators
        }));
      },
      AssignmentExpression: function AssignmentExpression(path) {
        if (this.inactive()) return; // Desugar assignment expressions so the visitor
        // can implement operator overloading

        var operator = {
          "+=": "+",
          "-=": "-",
          "*=": "*",
          "/=": "/",
          "%=": "%",
          "<<=": "<<",
          ">>=": ">>",
          ">>>=": ">>>",
          "|=": "|",
          "^=": "^",
          "&=": "&"
        }[path.node.operator];
        if (operator === undefined) return;

        var newValue = _core.types.BinaryOperator(operator, path.node.left, path.node.right);

        path.node.right = newValue;
        path.node.operator = "=";
      }
    }
  };
});

exports.default = _default;